/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "DropFilesDataObject.h"
#include <shlobj.h>
#include "DbgPrintf.h"
#include "DragDropGuest.h"
#include "WinTools.h"

const FORMATETC g_SupportedFormatEtc = {
		CF_HDROP, NULL,
		DVASPECT_CONTENT, -1, TYMED_HGLOBAL
		};



/*string
ClipboardFormatToName( UINT cfFormat )
{

	if ( cfFormat == CF_HDROP )
	{
		return "CF_HDROP";
	}

	string result;
	CHAR szFormatName[200] = "";

	GetClipboardFormatName(cfFormat, szFormatName, ARRAYSIZE(szFormatName) );

	result = szFormatName;

	return "";
}
*/


//===================		

STDMETHODIMP CDropFilesDataObject::QueryInterface(
	REFIID riid,
	void **ppv
	)
{

	if ( riid == IID_IUnknown )
	{
		*ppv = (IUnknown *)this;
		AddRef();
		return S_OK;
	}
	else if ( riid == IID_IDataObject )
	{
		*ppv = (IDataObject *)this;
		AddRef();
		return S_OK;
	}

	*ppv = NULL;
	return E_NOINTERFACE;
}

STDMETHODIMP_(ULONG) CDropFilesDataObject::AddRef()
{
  return InterlockedIncrement( &m_cRef );
}

STDMETHODIMP_(ULONG) CDropFilesDataObject::Release()
{
	LONG cRef = 0;

	if ( (cRef=InterlockedDecrement(&m_cRef)) == 0 )
	{		
        DbgPrintf( "CDropFilesDataObject destroy\n" );
		delete this;
	}

	return cRef;
}

STDMETHODIMP CDropFilesDataObject::GetData(
	FORMATETC *pformatetcIn,
	STGMEDIUM *pmedium
	)
{
    DbgPrintf( "CDropFilesDataObject::GetData\n" );

    return InternalGetData(
        pformatetcIn,
        pmedium
        );
}

HDROP
FileListToDropHandleW(
	vector<wstring> *pFileList,
	POINT pt
	)
{
	UINT BufferSize = sizeof(DROPFILES) + 2*sizeof(WCHAR);

    for (
		vector<wstring>::iterator iter = pFileList->begin();
        iter != pFileList->end();
        iter++
        )
    {
		BufferSize += ( (iter->length() + 1 ) * sizeof(WCHAR) );
    } 
    

    HGLOBAL hNewDropHandle = GlobalAlloc( 
        GPTR, //GHND | GMEM_SHARE, 
        BufferSize
        );

    if ( NULL == hNewDropHandle )
    {
		return NULL;
    }

    LPDROPFILES pDropFiles = (LPDROPFILES)GlobalLock( hNewDropHandle );

	if ( NULL == pDropFiles )
    {
        GlobalFree( hNewDropHandle );
        return NULL;
	}

    pDropFiles->fWide = TRUE;
	pDropFiles->pt = pt;
    pDropFiles->fNC = TRUE;
    pDropFiles->pFiles = sizeof( DROPFILES );

    PUCHAR pszCurr = (PUCHAR) ( (PUCHAR)pDropFiles + pDropFiles->pFiles  );

    for (
		vector<wstring>::iterator iter = pFileList->begin();
        iter != pFileList->end();
        iter++
        )
    {
        RtlCopyMemory(
            pszCurr,
            (PVOID)iter->c_str(),
            (iter->length()+1)*sizeof(WCHAR)
            );

        pszCurr += (iter->length()+1)*sizeof(WCHAR);
	}

	//adding 2 last null terminators

	RtlZeroMemory( pszCurr, sizeof(WCHAR)*2 );

	GlobalUnlock( hNewDropHandle );

	return (HDROP)hNewDropHandle;

}


HGLOBAL
_GlobalDuplicate(
	HGLOBAL hGlobal
	)
{
    if ( NULL == hGlobal )
        return NULL;

	UINT MemFlags = GlobalFlags( hGlobal );

	if ( GMEM_INVALID_HANDLE == MemFlags )
	{
    	return NULL;
	}

	SIZE_T MemSize =  GlobalSize( hGlobal );

	if ( MemSize == 0 )
	{
		return NULL;
	}

	BOOL bWasError = FALSE;
	HGLOBAL hNewGlobal = GlobalAlloc( MemFlags, MemSize );

	if ( hNewGlobal )
	{
		PVOID pBuffSrc = GlobalLock( hGlobal );

		if ( pBuffSrc )
		{
			PVOID pBuffDest = GlobalLock( hNewGlobal );

			if ( pBuffDest )
			{
				RtlCopyMemory( pBuffDest, pBuffSrc, MemSize );
				GlobalUnlock( hNewGlobal );
			}
			else
			{
            	bWasError = TRUE;
            }

			GlobalUnlock( hGlobal );
		}
		else
		{
			bWasError = TRUE;
		}
	}
	else
    	bWasError = TRUE;

	if ( bWasError )
	{
		if ( hNewGlobal )
		{
			GlobalFree( hNewGlobal );
			hNewGlobal = NULL;            
		}
	}

	return hNewGlobal;
}



STDMETHODIMP CDropFilesDataObject::GetDataHere(
	FORMATETC *pformatetc,
	STGMEDIUM *pmedium
	)
{
	DbgPrintf( "CDropFilesDataObject::GetDataHere\n" );

    return InternalGetData(
        pformatetc,
        pmedium
        );
}


STDMETHODIMP CDropFilesDataObject::QueryGetData(
	FORMATETC *pformatetc
	)
{
	HRESULT Result = S_OK;

	if ( !RtlEqualMemory( &g_SupportedFormatEtc, pformatetc, sizeof(g_SupportedFormatEtc) )  )
	{
		Result = DV_E_FORMATETC;
	}

	DbgPrintf( "QueryGetData(0x%x): cfFormat=%u: ptd=%p dwAspect=%u lindex=%d tymed=%u\n",
		(ULONG)Result,
		(ULONG)pformatetc->cfFormat,
		pformatetc->ptd,
		pformatetc->dwAspect,
		(ULONG)pformatetc->lindex,
		(ULONG)pformatetc->tymed
		);


	return Result;
}


STDMETHODIMP
CDropFilesDataObject::EnumFormatEtc(
	DWORD dwDirection,
	IEnumFORMATETC **ppenumFormatEtc
	)
{

	DbgPrintf( "CDropFilesDataObject::EnumFormatEtc\n" );

	if ( DATADIR_GET != dwDirection )
	{
		return E_NOTIMPL;
	}

	CDropFilesIEnumFORMATETC *pResult = new CDropFilesIEnumFORMATETC;

	if ( pResult )
	{
		*ppenumFormatEtc = pResult;
		return S_OK;        
	}

	return E_OUTOFMEMORY;
}


STDMETHODIMP CDropFilesIEnumFORMATETC::QueryInterface(REFIID riid, void **ppv)
{
	if ( riid == IID_IUnknown )
	{
		*ppv = (IUnknown *)this;
		AddRef();
		return S_OK;
	}
	else if ( riid == IID_IEnumFORMATETC )
	{
		*ppv = (IEnumFORMATETC *)this;
		AddRef();
		return S_OK;
	}

	*ppv = NULL;
	return E_NOINTERFACE;

}


STDMETHODIMP_(ULONG) CDropFilesIEnumFORMATETC::AddRef()
{
  return InterlockedIncrement( &m_cRef );
}

STDMETHODIMP_(ULONG) CDropFilesIEnumFORMATETC::Release()
{
	LONG cRef = 0;

	if ( (cRef=InterlockedDecrement(&m_cRef)) == 0 )
	{
        DbgPrintf( "CDropFilesIEnumFORMATETC destroy\n" );
		delete this;
	}

	return cRef;
}

STDMETHODIMP  CDropFilesIEnumFORMATETC::Next(
	ULONG celt,
	FORMATETC *rgelt,
	ULONG *pceltFetched
	)
{
	ULONG celFetched = 0;

	if ( m_Pos < 1 && celt >= 1 )
	{
		*rgelt = g_SupportedFormatEtc;
    	celFetched++;
		m_Pos++;
    }
		
	if ( pceltFetched )
	{
		*pceltFetched = celFetched;
	}

	return (celFetched == celt) ? S_OK : S_FALSE;
}

STDMETHODIMP  CDropFilesIEnumFORMATETC::Skip(
	ULONG celt
	)
{
	if ( celt == 1 && m_Pos == 0 ) {
		m_Pos = 1;
		return S_OK;
	}

	return S_FALSE;

}

STDMETHODIMP CDropFilesIEnumFORMATETC::Reset()
{
	m_Pos = 0;
	return S_OK;
}

STDMETHODIMP CDropFilesIEnumFORMATETC::Clone( IEnumFORMATETC **ppenum )
{
	IEnumFORMATETC *pResult = new CDropFilesIEnumFORMATETC( m_Pos );

	*ppenum = pResult;

	return pResult != NULL ? S_OK : E_OUTOFMEMORY;

}


STDMETHODIMP CDropFilesDataObject::InternalGetData(
	FORMATETC *pformatetc,
	STGMEDIUM *pmedium
	)
{
    if ( m_DropState == DROP_STATE_BEFORE_DROPPING )
    {
        return InternalGetDataBeforeDropping( pformatetc, pmedium );
    }
    else if ( m_DropState == DROP_STATE_DROP )
    {
        return InternalGetDataAfterDrop( pformatetc, pmedium );
    }
    else
    {
        DbgPrintf( "Wrong m_DropState value\n" );
        return E_FAIL;
    }

}

STDMETHODIMP CDropFilesDataObject::InternalGetDataBeforeDropping(
	FORMATETC *pformatetc,
	STGMEDIUM *pmedium
	)
{
	if ( !RtlEqualMemory( &g_SupportedFormatEtc, pformatetc, sizeof(g_SupportedFormatEtc) )  )
	{
		return DV_E_FORMATETC;
	}

	STGMEDIUM StgMediumTmp;

    RtlZeroMemory( &StgMediumTmp, sizeof(StgMediumTmp) );

	StgMediumTmp.hGlobal = _GlobalDuplicate( m_hDrop );

	if ( StgMediumTmp.hGlobal == NULL )
	{
		return E_OUTOFMEMORY;
	}

	StgMediumTmp.tymed = TYMED_HGLOBAL;
	StgMediumTmp.pUnkForRelease = NULL;

	*pmedium = StgMediumTmp;
	return S_OK;
}




STDMETHODIMP CDropFilesDataObject::InternalGetDataAfterDrop(
	FORMATETC *pformatetcIn,
	STGMEDIUM *pmedium
	)
{

    //FIXED

	if ( !RtlEqualMemory( &g_SupportedFormatEtc, pformatetcIn, sizeof(g_SupportedFormatEtc) )  )
	{
		return DV_E_FORMATETC;
	}

    if ( !m_DroppedFilesWasCopied )
    {
        m_DroppedFilesWasCopied = TRUE;

        DragDrop_CopyFilesToFolder( m_uDropHandle, m_DropFilesFolder.c_str() );

        //vector<wstring> FilesList;

        //GetDirFilesListW( m_DropFilesFolder.c_str(), &m_DroppedFilesList );
    }

    /*if ( m_DroppedFilesList.size() == 0 )
    {
        DbgPrintf( "m_DroppedFilesList is empty. Temp folder (%ws) does not contain files\n", m_DropFilesFolder.c_str() );
        return E_FAIL;
    }*/

	STGMEDIUM StgMediumTmp;

    RtlZeroMemory( &StgMediumTmp, sizeof(StgMediumTmp) );

	StgMediumTmp.hGlobal = _GlobalDuplicate( m_hDrop );

	if ( StgMediumTmp.hGlobal == NULL )
	{
		return E_OUTOFMEMORY;
	}

	StgMediumTmp.tymed = TYMED_HGLOBAL;
	StgMediumTmp.pUnkForRelease = NULL;

	*pmedium = StgMediumTmp;
	return S_OK;
}

STDMETHODIMP CDropSource::QueryInterface(REFIID riid, void **ppv)
{

	if ( riid == IID_IUnknown )
	{
		*ppv = (IUnknown *)this;
		AddRef();
		return S_OK;
	}
	else if ( riid == IID_IDropSource )
	{
		*ppv = (IDropSource *)this;
		AddRef();
		return S_OK;
	}

	*ppv = NULL;
	return E_NOINTERFACE;
}

STDMETHODIMP_(ULONG) CDropSource::AddRef()
{
  return InterlockedIncrement( &m_cRef );
}

STDMETHODIMP_(ULONG) CDropSource::Release()
{
    LONG cRef = InterlockedDecrement( &m_cRef );

    if ( cRef == 0 )
    {
        DbgPrintf( "CDropSource destroy\n" );
        delete this;
    }

    return cRef;
}

STDMETHODIMP CDropSource::QueryContinueDrag(
	BOOL fEscapePressed,
	DWORD grfKeyState
	)
{
    if ( m_Count++ < 1 )
	{
        DbgPrintf( "QueryContinueDrag: Continue drag\n" );
    	return S_OK;    
	}

	DbgPrintf( "Change state to CDropFilesDataObject::DROP_STATE_DROP\n" );

    m_pDropFilesDataObject->SetDropState(
        CDropFilesDataObject::DROP_STATE_DROP
        );

	return DRAGDROP_S_DROP;
}

STDMETHODIMP CDropSource::GiveFeedback(DWORD dwEffect)
{
  return DRAGDROP_S_USEDEFAULTCURSORS;
}

CDropFilesDataObject::CDropFilesDataObject( 
    UINT uDropHandle,
	vector<wstring> *pPreDroppingFileList,
	POINT pt,
    LPCWSTR lpszDropFolder
    ) : m_cRef(1), m_uDropHandle(uDropHandle) 
{ 
    m_DropState = DROP_STATE_BEFORE_DROPPING; 
    m_DroppedFilesWasCopied = FALSE; 

	m_PreDroppingFilesList = *pPreDroppingFileList;
    m_Point = pt;
    m_DropFilesFolder = lpszDropFolder;

    m_hDrop = FileListToDropHandleW(
        &m_PreDroppingFilesList,
        m_Point
        );

};

CDropFilesDataObject::~CDropFilesDataObject()
{
    if ( m_hDrop )
    {
        GlobalFree( m_hDrop );
        m_hDrop = NULL;
    }
}

VOID
CDropFilesDataObject::Test_SetDroppedFileList( 
    vector<wstring> *pFileList 
    )
{
    if ( m_hDrop )
    {
        GlobalFree( m_hDrop );
        m_hDrop = NULL;
    }

    m_hDrop = FileListToDropHandleW(
        pFileList,
        m_Point
        );

    m_PreDroppingFilesList = *pFileList;
}

