/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#ifndef DropFilesDataObjectH
#define DropFilesDataObjectH

#include <windows.h>
#include <string>
#include <vector>
#include "DbgPrintf.h"
using namespace std;

HDROP
FileListToDropHandleW(
	vector<wstring> *pFileList,
	POINT pt
	);

HGLOBAL
_GlobalDuplicate(
	HGLOBAL hGlobal
	);	


class CDropFilesIEnumFORMATETC : public IEnumFORMATETC
{
public:

	// *** IUnknown ***

	STDMETHODIMP QueryInterface(REFIID riid, void **ppv);
	STDMETHODIMP_(ULONG) AddRef();
	STDMETHODIMP_(ULONG) Release();

	// *** IEnumFORMATETC ***

	STDMETHODIMP  Next(
		ULONG celt,
		FORMATETC *rgelt,
		ULONG *pceltFetched
		);

	STDMETHODIMP  Skip(
			ULONG celt);

	STDMETHODIMP Reset();

	STDMETHODIMP Clone( IEnumFORMATETC **ppenum );


public:

	CDropFilesIEnumFORMATETC() : m_cRef(1) { m_Pos = 0; DbgPrintf( "CDropFilesIEnumFORMATETC\n" ); }
	CDropFilesIEnumFORMATETC( UINT Pos ) : m_cRef(1) { m_Pos = Pos; }

private:

	LONG m_cRef;
	UINT m_Pos;

}; 

class CDropFilesDataObject : public IDataObject
{

public:

	// *** IUnknown ***

    STDMETHODIMP QueryInterface(REFIID riid, void **ppv);
    STDMETHODIMP_(ULONG) AddRef();
	STDMETHODIMP_(ULONG) Release();

	// *** IDataObject ***

	STDMETHODIMP GetData(
		FORMATETC *pformatetcIn,
		STGMEDIUM *pmedium
		);

	STDMETHODIMP GetDataHere(
		FORMATETC *pformatetc,
		STGMEDIUM *pmedium
		);

	STDMETHODIMP QueryGetData(
		FORMATETC *pformatetc
		);


	STDMETHODIMP GetCanonicalFormatEtc(
		FORMATETC *pformatectIn,
		FORMATETC *pformatetcOut
		) { DbgPrintf( "GetCanonicalFormatEtc<not impl>\n" ); return DATA_S_SAMEFORMATETC; };

	STDMETHODIMP SetData(
		FORMATETC *pformatetc,
		STGMEDIUM *pmedium,
		BOOL fRelease
		){ DbgPrintf( "SetData<not impl>\n" ); return E_NOTIMPL; };

	STDMETHODIMP EnumFormatEtc(
			DWORD dwDirection,
			IEnumFORMATETC **ppenumFormatEtc
			);

	STDMETHODIMP DAdvise(
			FORMATETC *pformatetc,
			DWORD advf,
			IAdviseSink *pAdvSink,
			DWORD *pdwConnection
			){ DbgPrintf( "DAdvise<not impl>\n" ); return E_NOTIMPL; }

	STDMETHODIMP DUnadvise(
		DWORD dwConnection
		) { DbgPrintf( "DUnadvise<not impl>\n" ); return E_NOTIMPL; }

	STDMETHODIMP EnumDAdvise(
		IEnumSTATDATA **ppenumAdvise
		){ DbgPrintf( "EnumDAdvise<not impl>\n" ); return OLE_E_ADVISENOTSUPPORTED; }

public:

typedef enum _DROP_STATE
{
    DROP_STATE_BEFORE_DROPPING,
    DROP_STATE_DROP
} DROP_STATE;


public:

	CDropFilesDataObject(
        UINT uDropHandle,
    	vector<wstring> *pPreDroppingFileList,
    	POINT pt,
        LPCWSTR lpszDropFolder
        );

    ~CDropFilesDataObject();

    VOID
    Test_SetDroppedFilesWasCopied( BOOL bSet ) { m_DroppedFilesWasCopied = bSet; };

    VOID
    Test_SetDroppedFileList( vector<wstring> *pFileList  );


    VOID
    SetDropState( DROP_STATE DropState ){ m_DropState = DropState; };

	STDMETHODIMP InternalGetData(
		FORMATETC *pformatetcIn,
		STGMEDIUM *pmedium
		);

	STDMETHODIMP InternalGetDataBeforeDropping(
		FORMATETC *pformatetcIn,
		STGMEDIUM *pmedium
		);

	STDMETHODIMP InternalGetDataAfterDrop(
		FORMATETC *pformatetcIn,
		STGMEDIUM *pmedium
		);



private:

  LONG m_cRef;
  vector<wstring> m_PreDroppingFilesList;
  vector<wstring> m_DroppedFilesList;
  wstring m_DropFilesFolder;
  BOOL m_DroppedFilesWasCopied;
  POINT m_Point;
  DROP_STATE m_DropState;
  UINT m_uDropHandle;
  HDROP m_hDrop;

}; //CDropFilesDataObject


class CDropSource : public IDropSource
{
public:
  // *** IUnknown ***
  STDMETHODIMP QueryInterface(REFIID riid, void **ppv);
  STDMETHODIMP_(ULONG) AddRef();
  STDMETHODIMP_(ULONG) Release();

  // *** IDropSource ***
  STDMETHODIMP QueryContinueDrag(BOOL fEscapePressed, DWORD grfKeyState);
  STDMETHODIMP GiveFeedback(DWORD dwEffect);

public:

  CDropSource( CDropFilesDataObject *pDropFilesDataObject ) : m_cRef(1), m_Count(0) 
  {
     m_pDropFilesDataObject = pDropFilesDataObject;
     m_pDropFilesDataObject->AddRef();
  }
  ~CDropSource()
  {
    if ( m_pDropFilesDataObject )
    {
        m_pDropFilesDataObject->Release();
        m_pDropFilesDataObject = NULL;
    }
  }

private:
  LONG m_cRef;
  ULONG m_Count;
  CDropFilesDataObject *m_pDropFilesDataObject;

};


#endif //DropFilesDataObjectH



