/*
    Copyright (C) 2012 NewestTools.com(AlexLe)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.

*/

#include "ComUtil.h"

#ifdef _DEBUG
#include <stdio.h>
#endif

HRESULT
CreateInstanceWithProgID(
	IN LPCOLESTR lpszProgID,
	IN DWORD dwClsContext,
	OUT IDispatch **ppDispatch 
	)
{

	CLSID clsidApp;

	HRESULT hr = CLSIDFromProgID( lpszProgID, &clsidApp );

	if ( !SUCCEEDED( hr ) )
		return hr;

	IDispatch *pdispApp = NULL;

	hr = CoCreateInstance(
		clsidApp, 
		NULL, 
		CLSCTX_LOCAL_SERVER,
		IID_IDispatch,
		(void **)&pdispApp
		);

	if ( !SUCCEEDED( hr ) )
		return hr;

	*ppDispatch = pdispApp;
	return hr;
}


HRESULT
Disp_PropGet(
	OUT VARIANT *pvResult,
	IN IDispatch *pDisp,
	IN LPCOLESTR lpPropName
	)
{

	DISPID PropDispID;

	HRESULT hr = pDisp->GetIDsOfNames(
		IID_NULL, 
		( LPOLESTR *)&lpPropName,
		1, 
		LOCALE_SYSTEM_DEFAULT,
		&PropDispID
		);

	if ( !SUCCEEDED( hr ) )
		return hr;

	DISPPARAMS DispParams = { NULL, NULL, 0, 0 };

    DispParams.cArgs = 0;
    DispParams.rgvarg = NULL;

	hr = pDisp->Invoke(
		PropDispID,
		IID_NULL, 
		LOCALE_SYSTEM_DEFAULT,
		DISPATCH_PROPERTYGET,
		&DispParams,
		pvResult,
		NULL,
		NULL
		);

	return hr;
}


HRESULT
Disp_CallMethod(
	OUT VARIANT *pvResult OPTIONAL,
	IN IDispatch *pDisp,
	IN LPCOLESTR lpMethodName,
	IN ULONG ArgCount,
	...
	)
{
	DISPID PropDispID;

	HRESULT hr = pDisp->GetIDsOfNames(
		IID_NULL, 
		(LPOLESTR *)&lpMethodName,
		1, 
		LOCALE_SYSTEM_DEFAULT,
		&PropDispID
		);

	if ( !SUCCEEDED( hr ) )
		return hr;

	va_list vaList;

	va_start( vaList, ArgCount );

	DISPPARAMS DispParams = { NULL, NULL, 0, 0 };

    DispParams.cArgs = ArgCount;

	if ( ArgCount > 0 )
	{
		VARIANT *pVarArgs = (VARIANT *)malloc(
			sizeof(VARIANT)*ArgCount
			);

		if ( pVarArgs == NULL  )
			return E_OUTOFMEMORY;

		//FIXME: reverse order?
		for ( int u = ArgCount - 1; u >= 0; u-- )
		{
			VARIANT *pCurrentArg = va_arg( vaList, VARIANT *);
			pVarArgs[u] = *pCurrentArg;
		}

		DispParams.rgvarg = pVarArgs;
			//(VARIANTARG *)(PUCHAR(&ArgCount) + sizeof(PVOID));
	}

	unsigned int ErrArg = -1;

	hr = pDisp->Invoke(
		PropDispID,
		IID_NULL, 
		LOCALE_SYSTEM_DEFAULT,
		DISPATCH_METHOD,
		&DispParams,
		pvResult,
		NULL,
		&ErrArg
		);

#ifdef _DEBUG
	if ( DISP_E_TYPEMISMATCH == hr )
	{
		printf( "Disp_CallMethod: invalid arg %u\n", ErrArg );
	}
#endif


	if ( DispParams.rgvarg )
	{
		free( DispParams.rgvarg );
		DispParams.rgvarg = NULL;
	}

	va_end( vaList );

	return hr;
}

HRESULT
Disp_PropSet(
	OUT VARIANT *pvResult OPTIONAL,
	IN IDispatch *pDisp,
	IN LPCOLESTR lpPropName,
	IN VARIANT *pPropValue
	)
{
	DISPID PropDispID;

	HRESULT hr = pDisp->GetIDsOfNames(
		IID_NULL, 
		(LPOLESTR *)&lpPropName,
		1, 
		LOCALE_SYSTEM_DEFAULT,
		&PropDispID
		);

	if ( !SUCCEEDED( hr ) )
		return hr;

	DISPPARAMS DispParams = { NULL, NULL, 0, 0 };

    DispParams.cArgs = 1;
	DispParams.rgvarg = pPropValue;

	//for DISPATCH_PROPERTYPUT

	DISPID DispidNamed = DISPID_PROPERTYPUT;
	DispParams.cNamedArgs = 1;
    DispParams.rgdispidNamedArgs = &DispidNamed;

	unsigned int ErrArg = -1;

	hr = pDisp->Invoke(
		PropDispID,
		IID_NULL, 
		LOCALE_SYSTEM_DEFAULT,
		DISPATCH_PROPERTYPUT,
		&DispParams,
		pvResult,
		NULL,
		&ErrArg
		);

#ifdef _DEBUG
	if ( DISP_E_TYPEMISMATCH == hr )
	{
		printf( "Disp_PropSet: invalid arg %u\n", ErrArg );
	}
#endif


	return hr;

}

HRESULT
Disp_PropSetEx(
	OUT VARIANT *pvResult OPTIONAL,
	IN IDispatch *pDisp,
	IN LPCOLESTR lpPropName,
	IN ULONG ArgCount,
	...
	)
{
	DISPID PropDispID;

	HRESULT hr = pDisp->GetIDsOfNames(
		IID_NULL, 
		(LPOLESTR *)&lpPropName,
		1, 
		LOCALE_SYSTEM_DEFAULT,
		&PropDispID
		);

	if ( !SUCCEEDED( hr ) )
		return hr;

	va_list vaList;

	va_start( vaList, ArgCount );

	DISPPARAMS DispParams = { NULL, NULL, 0, 0 };

    DispParams.cArgs = ArgCount;

	if ( ArgCount > 0 )
	{
		VARIANT *pVarArgs = (VARIANT *)malloc(
			sizeof(VARIANT)*ArgCount
			);

		if ( pVarArgs == NULL  )
			return E_OUTOFMEMORY;

		//FIXME: reverse order?
		for ( int u = ArgCount - 1; u >= 0; u-- )
		{
			VARIANT *pCurrentArg = va_arg( vaList, VARIANT *);
			pVarArgs[u] = *pCurrentArg;
		}

		DispParams.rgvarg = pVarArgs;
			//(VARIANTARG *)(PUCHAR(&ArgCount) + sizeof(PVOID));
	}

	//for DISPATCH_PROPERTYPUT

	DISPID DispidNamed = DISPID_PROPERTYPUT;
	DispParams.cNamedArgs = 1;
    DispParams.rgdispidNamedArgs = &DispidNamed;


	unsigned int ErrArg = -1;

	hr = pDisp->Invoke(
		PropDispID,
		IID_NULL, 
		LOCALE_SYSTEM_DEFAULT,
		DISPATCH_PROPERTYPUT,
		&DispParams,
		pvResult,
		NULL,
		&ErrArg
		);

#ifdef _DEBUG
	if ( DISP_E_TYPEMISMATCH == hr )
	{
		printf( "Disp_PropSetEx: invalid arg %u\n", ErrArg );
	}
#endif


	if ( DispParams.rgvarg )
	{
		free( DispParams.rgvarg );
		DispParams.rgvarg = NULL;
	}

	va_end( vaList );

	return hr;
}

HRESULT
Disp_PropGetEx(
	OUT VARIANT *pvResult,
	IN IDispatch *pDisp,
	IN LPCOLESTR lpPropName,
	IN ULONG ArgCount,
	...
	)
{
	DISPID PropDispID;

	HRESULT hr = pDisp->GetIDsOfNames(
		IID_NULL, 
		(LPOLESTR *)&lpPropName,
		1, 
		LOCALE_SYSTEM_DEFAULT,
		&PropDispID
		);

	if ( !SUCCEEDED( hr ) )
		return hr;

	va_list vaList;

	va_start( vaList, ArgCount );

	DISPPARAMS DispParams = { NULL, NULL, 0, 0 };

    DispParams.cArgs = ArgCount;

	if ( ArgCount > 0 )
	{
		VARIANT *pVarArgs = (VARIANT *)malloc(
			sizeof(VARIANT)*ArgCount
			);

		if ( pVarArgs == NULL  )
			return E_OUTOFMEMORY;

		//FIXME: reverse order?
		for ( int u = ArgCount - 1; u >= 0; u-- )
		{
			VARIANT *pCurrentArg = va_arg( vaList, VARIANT *);
			pVarArgs[u] = *pCurrentArg;
		}

		DispParams.rgvarg = pVarArgs;
			//(VARIANTARG *)(PUCHAR(&ArgCount) + sizeof(PVOID));
	}

	//for DISPATCH_PROPERTYPUT

	unsigned int ErrArg = -1;

	hr = pDisp->Invoke(
		PropDispID,
		IID_NULL, 
		LOCALE_SYSTEM_DEFAULT,
		DISPATCH_PROPERTYGET,
		&DispParams,
		pvResult,
		NULL,
		&ErrArg
		);

#ifdef _DEBUG
	if ( DISP_E_TYPEMISMATCH == hr )
	{
		printf( "Disp_PropGetEx: invalid arg %u\n", ErrArg );
	}
#endif


	if ( DispParams.rgvarg )
	{
		free( DispParams.rgvarg );
		DispParams.rgvarg = NULL;
	}

	va_end( vaList );

	return hr;

}

